import {BG_OUTER} from "./colors.ts";
import {D, I} from "./engine/public.ts";
import {IGame, Point, Size} from "./engine/datatypes.ts";
import {getHuntMode} from "./huntmode.ts";
import {getPageLocation, Page, withCamera} from "./layout.ts";
import {getHud} from "./hud.ts";
import {getHotbar, Hotbar} from "./hotbar.ts";
import {getSkillsModal, SkillsModal} from "./skillsmodal.ts";
import {getSleepModal, SleepModal} from "./sleepmodal.ts";
import {getEndgameModal} from "./vnmodal.ts";

class MenuCamera {
  // measured in whole screens
  position: Point;
  target: Point;

  constructor({position, target}: {position: Point, target: Point}) {
    this.position = position;
    this.target = target;
  }

  update() {
    let adjust = (x0: number, x1: number) => {
      if (Math.abs(x1 - x0) < 0.01) { return x1; }
      return (x0 * 8 + x1 * 2) / 10;
    }
    this.position = new Point(
      adjust(this.position.x, this.target.x),
      adjust(this.position.y, this.target.y),
    );
  }
}
export class Game implements IGame {
  camera: MenuCamera;
  page: Page;
  #bottomThing: SkillsModal | SleepModal | Hotbar | null;

  constructor() {
    this.camera = new MenuCamera({
      position: new Point(0, 0),
      target: new Point(0, 0),
    });
    this.page = "Gameplay";

    this.#bottomThing = null;
  }

  update() {
    if (I.isKeyPressed("w")) {
      this.page = "Gameplay"
    }
    if (I.isKeyPressed("s")) {
      this.page = "Thralls"
    }

    this.camera.target = getPageLocation(this.page);
    D.camera = new Point(
      D.size.w * this.camera.position.x,
      D.size.h * this.camera.position.y,
    )
    this.camera.update();

    // state-specific updates
    this.updateGameplay();
  }

  draw() {
    // draw screen background
    let oldCamera = D.camera;
    D.camera = new Point(0, 0);
    D.fillRect(new Point(0, 0), D.size, BG_OUTER);
    D.camera = oldCamera;

    this.drawGameplay();

    // we draw all states at once and pan between them
    // mainFont.drawText({ctx: ctx, text: "You have been given a gift.", x: 0, y: 0})
    let mouse = I.mousePosition?.offset(D.camera);
    if (mouse != null) {
      D.invertRect(mouse.offset(new Point(-1, -1)), new Size(3, 3))
    }
  }

  updateGameplay() {
    this.#chooseBottomThing();

    withCamera("Gameplay", () => {
      getHuntMode().update();
    });
    withCamera("HUD", () => { getHud().update() })
    this.#bottomThing?.update();
    getEndgameModal().update();
  }

  drawGameplay() {
    withCamera("Gameplay", () => {
      getHuntMode().draw();
    });
    withCamera("HUD", () => { getHud().draw() })
    this.#bottomThing?.draw()
    getEndgameModal().draw();
  }

  #chooseBottomThing() {
    // This is explicitly chosen because we would prefer that
    // the same bottomThing be used in both draw and update,
    // meaning that events after this in updateGameplay should not affect
    // its value

    let skillsModal = getSkillsModal();
    if (skillsModal.isShown) {
      this.#bottomThing = skillsModal;
      return;
    }

    // between skillsModal and sleepModal, skillsModal wins
    // this is important because skillsModal can be
    // activated from sleepModal
    let sleepModal = getSleepModal();
    if (sleepModal.isShown) {
      this.#bottomThing = sleepModal;
      return;
    }

    // use the hotbar only as a matter of last resort
    this.#bottomThing = getHotbar();
  }

  // withCamera("Hotbar", () => { getHotbar().draw() })
}


export let game = new Game();